@@ -20,7 +20,8 @@ import re
 import string
 from collections.abc import Callable
 from enum import StrEnum
-from typing import Any, ClassVar, cast
+from typing import TYPE_CHECKING, Any, ClassVar, cast
+from uuid import UUID
 
 import httpx
 from bs4 import BeautifulSoup
@@ -40,11 +41,14 @@ from aviary.core import (
     eval_answer,
 )
 
+if TYPE_CHECKING:
+    from datasets import Dataset
+
 logger = logging.getLogger(__name__)
 
 
 # Jitter in case we have lots of requests at once, to space them out slightly
-@retry(stop=stop_after_attempt(2), wait=wait_exponential_jitter(initial=1, max=4))
+@retry(stop=stop_after_attempt(2), wait=wait_exponential_jitter(max=4))
 async def fetch_with_retry(
     client: httpx.AsyncClient, url: str, **kwargs
 ) -> httpx.Response:
@@ -73,8 +77,10 @@ class HotPotQAEnvState(BaseModel):
     )
     last_action_is_lookup: bool = Field(
         default=False,
-        description="Whether the last action was a lookup action."
-        "Default is False, as after reset the agent has not yet taken any action.",
+        description=(
+            "Whether the last action was a lookup action."
+            "Default is False, as after reset the agent has not yet taken any action."
+        ),
     )
     last_lookup: str | None = Field(
         default=None, description="The last lookup keyword."
@@ -180,6 +186,7 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
 
     def __init__(
         self,
+        question_id: UUID | None,
         question: str,
         correct_answer: str | float,
         correct_reward: float = 1.0,
@@ -189,6 +196,7 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
         proxy: str | None = None,
     ):
         super().__init__()
+        self.question_id = question_id
         self.question = question
         # Normalize the correct answer once here as a minor performance optimization
         self.normalized_correct_answer = normalize_answer(correct_answer)
@@ -214,7 +222,7 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
 
     @classmethod
     def from_task(cls, task: str) -> "HotPotQAEnv":
-        return cls(question=task, correct_answer=0.0)
+        return cls(question_id=None, question=task, correct_answer=0.0)
 
     async def calculate_answer_reward(self, answer: str | None) -> float:
         """Calculate the reward based on the agent's answer.
@@ -250,12 +258,19 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
                 - list[Tool]: A list of tools (search, lookup, and submit_answer) available for the agent.
 
         Example:
-            >>> env = HotPotQAEnv()
-            >>> initial_obs, tools = env.reset(seed=42, idx=5)
-            >>> print(initial_obs)
-            [Message(content='Question: <question_text>')]
-            >>> print(tools)
-            [<Tool: search>, <Tool: lookup>, <Tool: submit_answer>]
+            >>> env = HotPotQAEnv(
+            ...     question_id=UUID("00000000-5a8d-7341-5542-99441c6b9fe5"),
+            ...     question=(
+            ...         "Musician and satirist Allie Goertz wrote a song about the"
+            ...         ' "The Simpsons" character Milhouse, who Matt Groening named after who?'
+            ...     ),
+            ...     correct_answer="President Richard Nixon",
+            ... )
+            >>> obs, tools = await env.reset()
+            >>> print(obs)
+            [Message(role='user', content='Question: Musician and satirist Allie Goertz wrote a song about the "The Simpsons" character Milhouse, who Matt Groening named after who?')]
+            >>> print([t.info.name for t in tools])
+            ['search', 'lookup', 'submit_answer']
         """
         self.state = self.State()
         return [Message(content=f"Question: {self.question}")], self.tools
@@ -277,14 +292,17 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
                 - bool: The done status indicating whether the episode is finished.
 
         Example:
-            >>> env = HotPotQAEnv()
+            >>> from aviary.core import ToolCall
+            >>> env = HotPotQAEnv(
+            ...     question_id=UUID("00000000-5a8d-7341-5542-99441c6b9fe5"),
+            ...     question=(
+            ...         "Musician and satirist Allie Goertz wrote a song about the"
+            ...         ' "The Simpsons" character Milhouse, who Matt Groening named after who?'
+            ...     ),
+            ...     correct_answer="President Richard Nixon",
+            ... )
             >>> action = ToolRequestMessage(
-            ...     tool_calls=[
-            ...         ToolCall(
-            ...             function=ToolCallFunction(name="Search"),
-            ...             arguments={"entity": "Python"},
-            ...         )
-            ...     ]
+            ...     tool_calls=[ToolCall.from_name("search", entity="Python")]
             ... )
             >>> obs, done = await env.step(action)
             >>> print(obs, done)
@@ -309,17 +327,19 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
         # We accumulate reward across all tool calls in a step
         self.state.reward = 0.0
         response_messages = cast(
-            Messages,
-            # Ordered since things like search -> lookup need to be run in order.
+            "Messages",
+            # Non-concurrent since things like search -> lookup need to be run in order.
             # NOTE: Handling tool exceptions here keeps the trajectory going, but I don't
             # think the returned message is useful to the agent/learning. Disabling for now.
-            await self.exec_tool_calls(
-                action,
-                ordered=True,
-            ),
+            await self.exec_tool_calls(action, concurrency=False),
         )
         return response_messages, self.state.reward, self.state.done, False
 
+    async def get_id(self) -> str:
+        if self.question_id is None:
+            raise ValueError("No question ID was configured.")
+        return str(self.question_id)
+
     def export_frame(self) -> Frame:
         """Export the current state of the environment as a Frame object.
 
@@ -339,13 +359,22 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
                 - "answer" (Optional[str]): The answer provided by the agent, if any.
 
         Example:
-            >>> env = HotPotQAEnv()
+            >>> env = HotPotQAEnv(
+            ...     question_id=UUID("00000000-5a8d-7341-5542-99441c6b9fe5"),
+            ...     question=(
+            ...         "Musician and satirist Allie Goertz wrote a song about the"
+            ...         ' "The Simpsons" character Milhouse, who Matt Groening named after who?'
+            ...     ),
+            ...     correct_answer="President Richard Nixon",
+            ... )
             >>> frame = env.export_frame()
             >>> print(frame.info)
             {'steps': 0, 'done': False, 'reward': 0.0, 'answer': None}
         """
         return Frame(
             info={
+                "question_id": self.question_id,
+                "question": self.question,
                 "steps": self.state.steps,
                 "done": self.state.done,
                 "reward": self.state.reward,
@@ -411,7 +440,10 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
         if result_divs:  # mismatch
             result_titles = [clean_str(div.get_text().strip()) for div in result_divs]
             self.state.page = None
-            return f"Search for {entity!r} returned no results. Similar: {result_titles[:5]}."
+            return (
+                f"Search for {entity!r} returned no results. Similar:"
+                f" {result_titles[:5]}."
+            )
 
         page = [p.get_text().strip() for p in soup.find_all("p") + soup.find_all("ul")]
         if any("may refer to:" in p for p in page):  # Recurse
@@ -460,7 +492,6 @@ class HotPotQAEnv(Environment[HotPotQAEnvState]):
             'Milhouse is a regular target for school bully Nelson Muntz and his friends Jimbo Jones, Dolph Starbeam and
             Kearney Zzyzwicz.',
             'Milhouse has a crush on Bart's sister, Lisa, a common plot element.']
-
         """
         if not keyword:
             self.state.done = True
@@ -552,16 +583,17 @@ class HotPotQADataset(TaskDataset[HotPotQAEnv]):
     # SEE: https://huggingface.co/datasets/hotpotqa/hotpot_qa
     HOTPOTQA_HUGGING_FACE_DATASET = "hotpotqa/hotpot_qa"
 
-    def get_data_from_hugging_face(
-        self, split: str, hf_dataset: str = HOTPOTQA_HUGGING_FACE_DATASET
-    ) -> list[tuple[str, str]]:
-        """Convert a local file and split to a list of (question, answer) tuples."""
+    @staticmethod
+    def load_raw(
+        split: str, hf_dataset: str = HOTPOTQA_HUGGING_FACE_DATASET
+    ) -> "Dataset":
+        """Load the HotPotQA dataset split from Hugging Face."""
         if split in {"dev", "eval", "val"}:  # Map common aliases
             split = "validation"
         all_datasets = load_dataset(hf_dataset, name="fullwiki", trust_remote_code=True)
         try:
-            data = all_datasets[split].select_columns(
-                column_names=["question", "answer", "level"]
+            return all_datasets[split].select_columns(
+                column_names=["id", "question", "answer", "level"]
             )
         except KeyError as exc:
             raise ValueError(
@@ -569,15 +601,33 @@ class HotPotQADataset(TaskDataset[HotPotQAEnv]):
                 f" please specify a split from {set(all_datasets.keys())}."
             ) from exc
 
+    def get_data_from_hugging_face(
+        self, split: str, hf_dataset: str = HOTPOTQA_HUGGING_FACE_DATASET
+    ) -> list[tuple[UUID, str, str]]:
+        """Get a list of (id, question, answer) tuples for the Hugging Face dataset."""
+        data = self.load_raw(split, hf_dataset)
+
         if not all(  # Making up for datasets not being typed: https://github.com/huggingface/datasets/issues/3841
-            isinstance(d["question"], str) and isinstance(d["answer"], str | float)
+            isinstance(d["id"], str)
+            and isinstance(d["question"], str)
+            and isinstance(d["answer"], str | float)
             for d in data
         ):
             raise ValueError(
                 f"Dataset {hf_dataset!r} and split {split!r} contains invalid"
-                " question(s) or answer(s)."
+                " ID(s), question(s), or answer(s)."
             )
-        return [(d["question"], d["answer"]) for d in data if self._filter_task(d)]
+        return [
+            (
+                # Sadly, using UUID v4 doesn't work with left padding (has collisions)
+                # or right padding (is a lossy conversion), so leave version unspecified
+                UUID(bytes=b"\x00" * 4 + bytes.fromhex(d["id"])),  # Left zero-pad
+                d["question"],
+                d["answer"],
+            )
+            for d in data
+            if self._filter_task(d)
+        ]
 
     def __init__(
         self, split: str, config: HotPotQAEnvConfig | dict | None = None, **kwargs
